#  Copyright 2024 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from functools import wraps


def memo(param: str):
    def decorator(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            if not hasattr(self, param):
                setattr(self, param, func(self, *args, **kwargs))
            return getattr(self, param)

        return wrapper

    return decorator



#  Copyright 2023 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from synnax.hardware.writer import Writer
from synnax.hardware.retrieve import Retriever
from synnax.hardware.payload import Rack, Task, Device


class Client:
    __writer: Writer
    __retriever: Retriever

    def __init__(self, writer: Writer, retriever: Retriever) -> None:
        self.__writer = writer
        self.__retriever = retriever

    def create_rack(self, racks: list[Rack]) -> list[Rack]:
        return self.__writer.create_rack(racks)

    def delete_rack(self, keys: list[int]):
        self.__writer.delete_rack(keys)

    def create_task(self, tasks: list[Task]) -> list[Task]:
        return self.__writer.create_task(tasks)

    def delete_task(self, keys: list[int]):
        self.__writer.delete_task(keys)

    def create_device(self, devices: list[Device]) -> list[Device]:
        return self.__writer.create_device(devices)

    def delete_device(self, keys: list[str]):
        self.__writer.delete_device(keys)

    def retrieve_rack(self, keys: list[int] | None = None) -> list[Rack]:
        return self.__retriever.retrieve_rack(keys)

    def retrieve_task(
        self, rack: int | None = None, keys: list[int] | None = None
    ) -> list[Task]:
        return self.__retriever.retrieve_task(rack, keys)

    def retrieve_device(self, keys: list[str] | None = None) -> list[Device]:
        return self.__retriever.retrieve_device(keys)



#  Copyright 2023 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

import uuid

from freighter import Payload, UnaryClient

from synnax.util.normalize import normalize


class _GetRequest(Payload):
    range: uuid.UUID
    keys: list[str]


class _GetResponse(Payload):
    pairs: dict[str, str]


class _SetRequest(Payload):
    range: uuid.UUID
    pairs: dict[str, str]


class _DeleteRequest(Payload):
    range: uuid.UUID
    keys: list[str]


class _EmptyResponse(Payload):
    ...


class KV:
    __SET_ENDPOINT = "/range/kv/set"
    __GET_ENDPOINT = "/range/kv/get"
    __DELETE_ENDPOINT = "/range/kv/delete"

    __client: UnaryClient
    __rng: uuid.UUID

    def __init__(self, rng: uuid.UUID, client: UnaryClient) -> None:
        self.__client = client
        self.__rng = rng

    def get(self, keys: str) -> str:
        ...

    def get(self, keys: str | list[str]) -> dict[str, str]:
        req = _GetRequest(range=self.__rng, keys=normalize(keys))
        res, exc = self.__client.send(self.__GET_ENDPOINT, req, _GetResponse)
        if exc is not None:
            raise exc
        if isinstance(keys, str):
            return res.pairs[keys]
        return res.pairs

    def set(self, key: str, value: str):
        ...

    def set(self, key: dict[str, str]):
        ...

    def set(self, key: str | dict[str, str], value: str | None = None) -> None:
        if isinstance(key, str):
            if value is None:
                raise ValueError("value is required when providing a single key")
            key = {key: value}
        req = _SetRequest(range=self.__rng, pairs=key)
        res, exc = self.__client.send(self.__SET_ENDPOINT, req, _EmptyResponse)
        if exc is not None:
            raise exc

    def delete(self, keys: str | list[str]) -> None:
        req = _DeleteRequest(range=self.__rng, keys=normalize(keys))
        res, exc = self.__client.send(self.__DELETE_ENDPOINT, req, _EmptyResponse)
        if exc is not None:
            raise exc

    # Implement dict-like interface
    def __getitem__(self, key: str) -> str:
        return self.get(key)

    def __setitem__(self, key: str, value: str) -> None:
        self.set(key, value)

    def __delitem__(self, key: str) -> None:
        self.delete(key)



#  Copyright 2023 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from freighter import Payload, UnaryClient
from synnax.hardware.payload import Rack, Task, Device
from alamos import NOOP, Instrumentation, trace


class _RetrieveTaskRequest(Payload):
    rack: int | None = None
    keys: list[int] | None = None


class _RetrieveTaskResponse(Payload):
    tasks: list[Task] | None = None


class _RetrieveDeviceRequest(Payload):
    keys: list[str] | None = None


class _RetrieveDeviceResponse(Payload):
    devices: list[Device] | None = None


class _RetrieveRackRequest(Payload):
    keys: list[int] | None = None


class _RetrieveRackResponse(Payload):
    racks: list[Rack] | None = None


RETRIEVE_TASK_ENDPOINT = "/hardware/task/retrieve"
RETRIEVE_DEVICE_ENDPOINT = "/hardware/device/retrieve"
RETRIEVE_RACK_ENDPOINT = "/hardware/rack/retrieve"


class Retriever:
    __client: UnaryClient
    instrumentation: Instrumentation = NOOP

    def __init__(
        self,
        client: UnaryClient,
        instrumentation: Instrumentation = NOOP,
    ) -> None:
        self.__client = client
        self.instrumentation = instrumentation

    @trace("debug")
    def retrieve_task(
        self,
        rack: int | None = None,
        keys: list[int] | None = None,
    ) -> list[Task]:
        res, exc = self.__client.send(
            RETRIEVE_TASK_ENDPOINT,
            _RetrieveTaskRequest(rack=rack, keys=keys),
            _RetrieveTaskResponse,
        )
        if exc is not None:
            raise exc
        if res.tasks is None:
            return list()
        return res.tasks

    @trace("debug")
    def retrieve_device(self, keys: list[str] | None = None) -> list[Device]:
        res, exc = self.__client.send(
            RETRIEVE_DEVICE_ENDPOINT,
            _RetrieveDeviceRequest(keys=keys),
            _RetrieveDeviceResponse,
        )
        if exc is not None:
            raise exc
        if res.devices is None:
            return list()
        return res.devices

    @trace("debug")
    def retrieve_rack(self, keys: list[int] | None = None) -> list[Rack]:
        res, exc = self.__client.send(
            RETRIEVE_RACK_ENDPOINT,
            _RetrieveRackRequest(keys=keys),
            _RetrieveRackResponse,
        )
        if exc is not None:
            raise exc
        if res.racks is None:
            return list()
        return res.racks



#  Copyright 2023 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from synnax.channel.client import Channel, ChannelClient
from synnax.channel.payload import (
    ChannelKey,
    ChannelKeys,
    ChannelName,
    ChannelNames,
    ChannelParams,
    ChannelPayload,
    normalize_channel_params,
)
from synnax.channel.retrieve import ChannelRetriever



#  Copyright 2023 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.


from typing import Unpack

from synnax.cli.console.sugared import AskKwargs
from synnax.cli.flow import Context
from synnax.telem import DataType, TimeSpan, TimeSpanUnits


def select_data_type(
    ctx: Context,
    **kwargs: Unpack[AskKwargs[str]],
) -> DataType | None:
    """Prompts the user to select a data type from a list of all available data
    types.

    :param ctx: The current flow Context.
    :param allow_none: Whether to allow the user to select None.
    """
    return DataType(
        ctx.console.select(
            rows=[str(name) for name in DataType.ALL],
            type_=str,
            columns=["data_type"],
            **kwargs,
        )[0]
    )


def ask_time_units_select(
    ctx: Context,
    question: str | None = None,
    **kwargs: Unpack[AskKwargs[str]],
) -> TimeSpanUnits:
    """Prompts the user to select a time unit from a list of all available time
    units.

    :param ctx: The current flow Context.
    :param allow_none: Whether to allow the user to select None.
    """
    if question is not None:
        ctx.console.info(question)
    return ctx.console.select(
        rows=["iso", *list(TimeSpan.UNITS.keys())],
        type_=str,
        columns=["unit"],
        **kwargs,
    )[0]
